/*
*******************************************************************
*                                                                 *
*                         DiskSims 1.1 Beta                       *
*                    A Disk Schduling Simulator                   *
*       Released under the GPL (GNU General Public Licence)       *
*                                                                 *
*               Copyright (c) 2004-2008 Ammar Muqaddas            *
*                        ammar@solostuff.net                      *
*                        ammar.mu@gmail.com                       *
*                         www.solostuff.net                       *
*                                                                 *
*******************************************************************
*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <memory.h>
#include <math.h>
#include <windows.h>
#include "mt19937int.h"

// Enable/Disable rotational latency
#define RLATENCY

#define ADD_BPTTIME(bltime,at_time) ((bltime>=at_time) ? bltime : bltime + BPTTIME)

#ifdef RLATENCY
	#define NOWAIT(ALGO) if (ALGO##_insrv<BLOCKTIME) ALGO##_wait--
		
	// Service time equals block time + rotation time.
	#define SRVTIME(ALGO,qname) BLOCKTIME + ADD_BPTTIME(qname[ALGO##_nexti].bltime,cu_bltime) - cu_bltime
	
	#define MSG_NO_RLATENCY(opt)
#else
	#define NOWAIT(ALGO) ALGO##_wait--
		
	// Service time equals block time
	#define SRVTIME(ALGO,qname) BLOCKTIME 

	#define MSG_NO_RLATENCY(opt) printf("You should #define RLATENCY to use " #opt " option.\n"); 
#endif


typedef struct _qreq
	{	int cyl;		// Cylinder number
		#ifdef RLATENCY
		int bltime;		// (block number)*BLOCKTIME
		#endif
	} qreq;

struct qnode
	{	int reqi;
		int g,h,f,minsi,maxsi;
		int *pp;
		struct qnode *link;
	};

struct qsort_elm
	{	int index,req;
	};

const double GENRAND_MAX = 4294967295;   // 2^32-1
const long WM_PROGRESS = WM_APP + 1, WM_STOP = WM_APP + 2, WM_THREADID = WM_APP + 3;
int MAXCYLINDER = 1000, STARTCYL = 0, BLOCKTIME = 2, CYLSEEK = 1, BPT = -1, BPT_1 = -1;
#ifdef RLATENCY
	int cu_bltime, BPTTIME;
#endif
int total_requests=-1;
long caller_hwnd = -1, chunk = -1, chunk_count = 0;
unsigned long peek_count = 0;

int (*get_SSTF_nexti)(qreq *q, int SSTF_last, int head);
int (*get_SCAN_nexti)(qreq *q, int SCAN_last, int head, int direc);


int get_SSTF_nexti_nor(qreq *q, int SSTF_last, int head)
{	int n, neari = total_requests; 
	
	for (n=0; n<=SSTF_last; n++)
		if (abs(head-q[n].cyl)<abs(head-q[neari].cyl)) neari=n; 

	if (neari==total_requests) neari=-1;
	return neari;
}

int get_SCAN_nexti_nor(qreq *q, int SCAN_last, int head, int direc)
{	int n,a, neari = total_requests;
	
	for (n=0; n<=SCAN_last; n++)
	{	a=(q[n].cyl-head);
		if (a*direc>=0 && abs(a)<abs(q[neari].cyl-head)) neari=n;
	}
	if (neari==total_requests) neari=-1;
	return neari;
}

__inline void fix_nexti(qreq *q, int last,int head,int direc,int *pnexti)
{	// last is the index of the last request, which is the new request

	int a;
	a=(q[last].cyl-head);
	if ((*pnexti)>-1 && a*direc>=0 && abs(a)<abs(q[*pnexti].cyl-head)) *pnexti=last;
}


#ifdef RLATENCY
int (*get_cyl_nexti)(qreq *q, int last, int neari);

int get_cyl_nexti_rlat(qreq *q, int last, int neari)
{	// A neari is given to the function. Then it tries to get a nearer request to the head 
	// than neari according rotational latency. It doesn't find requests on other cylinders 
	// Note: To work correctly. the head has to be equal to q[neari].cyl. i.e the head is standing
	// on the cylinder of request neari

	int n, cyl; 

	cyl=q[neari].cyl;
	for (n=0; n<=last; n++)
		if (q[n].cyl==cyl && ADD_BPTTIME(q[n].bltime,cu_bltime) < ADD_BPTTIME(q[neari].bltime,cu_bltime)) neari=n;	//**
	
	return neari;
}

int get_cyl_nexti_same(qreq *q, int last, int neari)
{	return neari;
}

int get_SSTF_nexti_rlat(qreq *q, int SSTF_last, int head)
{	
	int n, neari = total_requests, a1; 
	for (n=0; n<=SSTF_last; n++)
	{	a1 = abs(head-q[n].cyl) - abs(head-q[neari].cyl);
		if (a1<0) neari=n; 
		else if (a1==0) 
		{	//int a2;
			a1 = (cu_bltime + abs(head-q[n].cyl) * CYLSEEK) % BPTTIME;
			if (ADD_BPTTIME(q[n].bltime,a1) < ADD_BPTTIME(q[neari].bltime,a1)) neari=n;	//**
		}
	}

	if (neari==total_requests) neari=-1;
	return neari;
}

int get_SCAN_nexti_rlat(qreq *q, int SCAN_last, int head, int direc)
{	int n,a, neari = total_requests, a1;
	for (n=0; n<=SCAN_last; n++)
	{	a=(q[n].cyl-head);
		if (a*direc>=0)
		{	a1 = abs(a) - abs(q[neari].cyl-head);
			if (a1<0) neari=n;
			else if (a1==0)
			{	//int a2;
				a1 = (cu_bltime + abs(a) * CYLSEEK) % BPTTIME;
				if (ADD_BPTTIME(q[n].bltime,a1) < ADD_BPTTIME(q[neari].bltime,a1)) neari=n;
			}
		}
	}
	if (neari==total_requests) neari=-1;
	return neari;
}

__inline short fix_nexti_rlat(qreq *q, int last,int head,int direc,int *pnexti, long headmv)
{	// Same as fix_nexti but with rotational latency enabled

	int a;
	a=(q[last].cyl-head);
	if ((*pnexti)>-1 && a*direc>=0)
		if (abs(a)<abs(q[*pnexti].cyl-head)) { *pnexti=last; return 1;}
		else if (abs(a)==abs(q[*pnexti].cyl-head))
		{	int a1;
			a1 = (cu_bltime + abs(a) * CYLSEEK + headmv) % BPTTIME;
			if (ADD_BPTTIME(q[last].bltime,a1) < ADD_BPTTIME(q[*pnexti].bltime,a1)) { *pnexti=last; return 2;}
		}
	return 0;
}
#endif

/* int *remove_sim(int *q, int last)
{	int *rq = malloc(sizeof(int)*(last+1));
	int n,n1,rlast=-1;
	short flag1=0;
	for (n=0; n<=last; n++)
	{	flag1=0;
		for (n1=0; n1<=rlast; n1++) flag1=(flag1 || q[n]==rq[n1]);
		if (!flag1) { rlast++; rq[rlast]=q[n] }
	}
} */

int compare(const struct qsort_elm *elem1,const struct qsort_elm *elem2)		//** Definistion is void
{	return (elem1->req)-(elem2->req);
}

void print_listf(struct qnode *plist)
{	while (plist!=NULL)
	{	printf(" %d",plist->f);
		plist=plist->link;
	}
}

void list_finsort(struct qnode **pplist, struct qnode *pitem)
{	if ((*pplist)!=NULL)
		while ((pitem->f)>(*pplist)->f)
		{	pplist=&((*pplist)->link);
			if ((*pplist)==NULL) break;
		}
	pitem->link=*pplist;
	(*pplist)=pitem;
}

__inline void update_progress_bar()
{	if (chunk != -1)
	{	chunk_count++;
		if (chunk_count>=chunk) { PostMessage((HWND) caller_hwnd,WM_PROGRESS,chunk_count,0); chunk_count=0; }
	}
}

void main(int argc, char *argv[])
{ 
	int max_req_pert=-1; 
	int rcount = 0, head=-1;
	int n,a,a1,a2,last = -1, qmlast=-1,qmi,qmi2;
	char c1, *wcolumn = NULL, *scolumn = NULL;
	long FCFS_head, FCFS_nexti = 0, FCFS_seek = 0, FCFS_headmv=0;
	long SSTF_head, SSTF_nexti = -1, SSTF_seek = 0, SSTF_cleft = 0, SSTF_last=last, SSTF_headmv=0;
	long SCAN_head, SCAN_nexti = -1, SCAN_seek = 0, SCAN_cleft = 0, SCAN_last=last, SCAN_direc=1, SCAN_headmv=0;
	long CSCAN_head, CSCAN_nexti = -1, CSCAN_seek = 0, CSCAN_cleft = 0, CSCAN_last=last, CSCAN_direc=1, CSCAN_headmv=0;
	long LOOK_head, LOOK_nexti = -1, LOOK_seek = 0, LOOK_cleft = 0, LOOK_last=last, LOOK_direc=1, LOOK_headmv=0;
	long CLOOK_head, CLOOK_nexti = -1, CLOOK_seek = 0, CLOOK_cleft = 0, CLOOK_last=last, CLOOK_direc=1, CLOOK_headmv=0;
	long SLOOK_head, SLOOK_nexti = -1, SLOOK_seek = 0, SLOOK_cleft = 0, SLOOK_last=last, SLOOK_direc=1, SLOOK_headmv=0, SLOOK_waist=0, SLOOK_wtfact=-1, SLOOK_wtconst=0; 
	long BOPATH_head, BOPATH_nexti = -1, BOPATH_seek = 0, BOPATH_cleft = 0, BOPATH_last=last, BOPATH_direc=1, BOPATH_headmv=0, BOPATH_optlast=-1, BOPATH_waist=0, BOPATH_wthresh=-1, BOPATH_maxbnd=0; //, BOPATH_mini, BOPATH_maxi,  BOPATH_wtfact; 


	__int64 FCFS_wait = 0, SSTF_wait = 0, SCAN_wait = 0, CLOOK_wait = 0, CSCAN_wait = 0, LOOK_wait = 0, SLOOK_wait = 0, BOPATH_wait = 0;
	short FCFS_insrv = 0, SSTF_insrv = 0, SCAN_insrv = 0, CSCAN_insrv = 0, LOOK_insrv = 0, CLOOK_insrv = 0, SLOOK_insrv = 0, SLOOK_mnmxfl = 0, SLOOK_flag1=-2, BOPATH_insrv = 0, BOPATH_wresetfl=-2, BOPATH_flag1=0, BOPATH_swap=-2, manual = 0, dislabel = 0, skip_idle=0, rlat_method=-1;
	double req_prob=-1, arrival_r=-1, time_unit=-1, poisson_e;   // poisson_e = (e^-a)
	FILE *wstream = NULL, *sstream = NULL, *istream=stdin;
	MSG msg1;

	qreq *q;
	qreq *q2;
	qreq *q3;
	qreq *q4;
	qreq *q5;
	qreq *q6;
	//int *qtemp;

	// Queus for manual request entry
	int *qm=NULL;
	int *qb=NULL;
	
	srand((unsigned)time(NULL));
	sgenrand((unsigned)time(NULL));
	//srand(600); sgenrand(800);

	for(n=0; n<argc; n++)
		switch (toupper(argv[n][1]))
		{	case 'C':
				if (toupper(argv[n][2])=='W') wcolumn = &argv[n][4];
				else if (toupper(argv[n][2])=='S') scolumn = &argv[n][4];
				else chunk=atoi(&(argv[n][3]));
				break;
			case 'W':
				if (toupper(argv[n][2])=='C') { SLOOK_wtconst = atoi(&(argv[n][4])); SLOOK_waist=SLOOK_wtconst; }
				else wstream = fopen(&(argv[n][3]),"a+");
				break;
			case 'S':
				sstream = fopen(&(argv[n][3]),"a+");
				break;
			case 'D':
				dislabel = -1;
				break;
			case 'B':
				if ((toupper(argv[n][2])=='P') && (toupper(argv[n][3])=='T'))
				{	MSG_NO_RLATENCY(-bpt);
					BPT = atoi(&(argv[n][5]));
				}
				else BLOCKTIME=atoi(&(argv[n][3]));
				break;
			case 'E':
				CYLSEEK=atoi(&(argv[n][3]));
				break;
			case 'M':
				if (toupper(argv[n][2])=='T') 
				{	MSG_NO_RLATENCY(-mt);
					rlat_method = (short) atoi(&(argv[n][4]));
				}
				else MAXCYLINDER=atoi(&(argv[n][3]))-1;
				break;
			case 'R':
				if (toupper(argv[n][2])=='W') SLOOK_flag1=-(toupper(argv[n][4])=='Y');
				else if (toupper(argv[n][2])=='B') BOPATH_wresetfl=-(toupper(argv[n][4])=='Y');
				else if (toupper(argv[n][2])=='S') BOPATH_swap=-(toupper(argv[n][4])=='Y');
				else { srand(atoi(&(argv[n][3]))); sgenrand(atoi(&(argv[n][3]))); }
				break;
			case 'T':
				if (toupper(argv[n][2])=='U') time_unit=atof(&(argv[n][4]));
				else { total_requests=atoi(&(argv[n][3])); manual=(total_requests==0); }
				break;
			case 'N':
				max_req_pert=atoi(&(argv[n][3]));
				break;
			case 'P':
				req_prob=atof(&(argv[n][3]));
				//else if (argv[n][2]=='2') req_prob2=atof(&(argv[n][4]));
				//else if (argv[n][2]=='3') req_prob3=atof(&(argv[n][4]));
				break;
			case 'I':
				if (toupper(argv[n][2])=='N') istream = fopen(&(argv[n][4]),"r");
				else if (toupper(argv[n][2])=='D') skip_idle=-1;
				else head=atoi(&(argv[n][3]));
				break;
			case 'F':
				if ((toupper(argv[n][2])=='C') && (toupper(argv[n][3])=='F') && (toupper(argv[n][4])=='S'))
					FCFS_nexti=-2;
				else SLOOK_wtfact=atoi(&(argv[n][3]));
				break;
			case 'O':
				BOPATH_wthresh=atoi(&(argv[n][3]));
				break;
			case 'X':
				BOPATH_maxbnd=atoi(&(argv[n][3]));
				break;
			case 'A':
				arrival_r = atof(&(argv[n][3]));
				break;
			case '?':
				; 
			case 'H':
				if (toupper(argv[n][2])=='W') 
				{	caller_hwnd = atoi(&(argv[n][4]));
					PostMessage((HWND) caller_hwnd,WM_THREADID,GetCurrentThreadId(),0);
				}
				else {
				printf("\nDiskSims v1.1 Beta. A Disk Scheduling Simulator\nCopyright (c) 2004-2008 Ammar Muqaddas\n\n");
				printf("Usage:  schedule [-w:waittime_file -cw:column] [-s:seek_file -cs:column] [-d] [-in:infile] [-b:blocktime] [-e:seektime] [-m:cylinders] [-bpt:bpt] [-r:seed] [-t:total] [-n:MNRPTU] [-a:rate -tu:time] [-p:probability] [-i:head] [-f:factor] [-rw:y/n] [-o:threshold] [-x:maxdepth] [-rb:(y/n)] [-rs:(y/n)] [-mt:(1/2)] [-hw:hwnd [-c:chunk]] [-id]\n\n");
				printf("   -w:waittime_file     Specifies an output csv file for waiting time results.\n");
				printf("   -s:seek_file         Specifies an output csv file for seek time results.\n");
				printf("   -cw:column           A value that will be writen to the first column of\n                        the next row in <waittime_file>.\n");
				printf("   -cs:column           A value that will be writen to the first column of\n                        the next row in <seek_file>.\n");
				printf("   -d                   Disable adding column labels in new or empty files.\n");
				printf("   -in:infile           Specifies an input file. The program will take it's\n                        input from this file instead of asking the user.\n");
				printf("   -b:blocktime         Single block transfer time (in time units)\n");
				printf("   -e:seektime          Head's single cylinder movement time (in time units)\n");
				printf("   -m:cylinders         Total number of cylinders in hard disk.\n");
				printf("   -bpt:bpt             Number of blocks per track. RLATENCY should be defined.\n");         
				printf("   -r:seed              Randomization seed. Ensures that same requests are\n                        generated on each run. If omitted, The timer is used\n                        which results in truly random request generation.\n");
				printf("   -t:total             Total number of requests to be generated.\n");
				printf("   -n:MNRPTU            Maximum number of requests per time unit.\n");
				printf("   -a:rate              Requests arrival rate per second. Used for Poisson                              Distribution. If used -p and -n are ignored.\n");
				printf("   -tu:time             Seconds per time unit. Used for Poisson Distribution.                           If used -p and -n are ignored.\n");
				printf("   -p:probability       Prbability of whether to generate requests or not\n                        in each time unit.\n");
				printf("   -i:head              Head's initial position. (cylinder number).\n");
				printf("   -f:factor            S-LOOK waisted time factor.\n");
				printf("   -wc:constant         S-LOOK waited time constant\n");				
				printf("   -rw:(y/n)            Should be 'yes' or 'no'. Indicates that the SLOOK\n                        waisted time should be reset on every minimum or                                maximum request.\n");
				printf("   -o:threshold         BO-PATH waisted time threshold.\n");
				printf("   -x:maxdepth	        BO-PATH maximum A* depth (Min=1).\n");
				printf("   -rb:(y/n)            Should be 'yes' or 'no'. Indicates that BO-PATH\n                        waisted time should be reset on every request in                                path.\n");
				printf("   -rs:(y/n)            Should be 'yes' or 'no'. Use BO-PATH min/max\n                        swaping trick.\n");
				printf("   -mt:(1/2)            Should be 1 or 2. Rotational Latency method for\n");
				printf("                        dealing with requests on the same cylinder.\n");
				//printf("                        the closest request immediately\n");
				printf("   -hw:hwnd             (For system use). Window handle of calling process.\n");
				printf("   -c:chunk             (For system use). Number of messages to collect\n                        before sending a chunked progress message.\n");
				printf("   -id                  Skip idle simulation time to improve performance.\n                        Idle time is when all heads of all algorithms \n                        are not doing any thing. Usefull for very low request\n                        probabilities.");
				exit(0); 
				}
		}
	if (arrival_r<=0 || time_unit<=0) { arrival_r=-1; time_unit=-1; }
	else max_req_pert=0;

	#ifdef RLATENCY
		// Setting some default values if they are not specified.
		if (BPT==-1) { BPT=600; BPT_1=599; }
		if (rlat_method==-1) rlat_method=1;

	#else
		// Checking for options that are NOT enabled by #defining compiler directives
		if (caller_hwnd==-1 && (BPT!=-1 || rlat_method!=-1))
			exit(0);
	#endif
		
	// Writing Column Labels
	if (!dislabel)
	{	if (wstream!=NULL)
		{	fseek(wstream,0,SEEK_END);
			if (ftell(wstream)==0) fprintf(wstream,",FCFS,SSTF,SCAN,C-SCAN,LOOK,C-LOOK,S-LOOK,BO-PATH\n");
		}
		if (sstream!=NULL)
		{	fseek(sstream,0,SEEK_END);
			if (ftell(sstream)==0) fprintf(sstream,",FCFS,SSTF,SCAN,C-SCAN,LOOK,C-LOOK,S-LOOK,BO-PATH\n");
		}
	}

	//printf("%s\n",argv[n]);


	// Entering values
	if (total_requests==-1)
	{	printf("\nEnter total number of requests (0=Manual Entry): ");
		fscanf(istream,"%d", &total_requests); manual=(total_requests==0);
	}
	if (manual)
	{	a=20; a1=-1; total_requests=0; qmi=0;
		qm=realloc(qm,a*sizeof(int));
		#ifdef RLATENCY
			qb=realloc(qb,a*sizeof(int));
		#endif

		// Begin manual entry loop
		do
		{	do
			{	a2=getc(istream);
				if (a2=='\n') 
				{	a1++; printf("Enter requests for t=%d (-1=stop): ",a1); qmlast++; 
					
					// Check if array size is enough and increase if not.
					if ((qmlast+1)>a) 
					{	a=a+20; qm=realloc(qm,a*sizeof(int)); 
						#ifdef RLATENCY
							qb=realloc(qb,a*sizeof(int)); 
						#endif
					}
					qm[qmlast]=-2; //qb[qmlast]=-2;
				}
			} while (a2==' ' || a2=='\n');
			//if (a2=='\r') printf("***");
			ungetc(a2,istream); qmlast++; 
			
			// Check if array size is enough and increase if not.
			if ((qmlast+1)>a) 
			{	a=a+20; qm=realloc(qm,a*sizeof(int)); 
				#ifdef RLATENCY
					qb=realloc(qb,a*sizeof(int)); 
				#endif
			}
			total_requests++; 
			
			// Get the cylinder number
			fscanf(istream,"%d",&qm[qmlast]); 
			
			#ifdef RLATENCY
				// Bypass a delimiter ':'
				a2=getc(istream);

				// Get the block nmuber
				fscanf(istream,"%d",&qb[qmlast]);
			#endif
		} while (qm[qmlast]!=-1);

		for (n=0; n<=qmlast; n++)
			printf(" %d",qm[n]);
		qmlast--; total_requests--;
		printf("\nTotal number of requests=%d\n\n",total_requests);
	} else
	{	if (max_req_pert==-1)
		{	printf("\nEnter Maximum number of requests per time unit (0=Generate All in t=0): ");
			fscanf(istream,"%d", &max_req_pert);
		}
		if (max_req_pert)
		{	if (req_prob==-1)
			{	printf("\nEnter request probability %%: ");		// Minimum non-zero probability is 1/(GENRAND_MAX+1)
				if (req_prob!=-1) printf("%g ", req_prob); 
				if (req_prob==-1) fscanf(istream," %lg", &req_prob); 
			}
			//if ((req_prob<0.00306 || req_prob2<0.00306 || req_prob3<0.00306) && req_prob!=0 && req_prob2!=0 && req_prob3!=0) 
			//	{ req_prob=0; printf(">> One of the probabilities is too low, Assumming 0%%\n"); 
			//	}
			req_prob = req_prob/100; printf("\n>> %.10f\n", req_prob);
		} else req_prob=0;
	}  // if Manual
	
	if (head==-1)
	{	printf("Enter head's initial position: ");
		fscanf(istream,"%d", &a);
	} else a=head;
	head=a; FCFS_head=a; SSTF_head=a; SCAN_head=a; CSCAN_head=a; LOOK_head=a; CLOOK_head=a; SLOOK_head=a; BOPATH_head=a;
	if (SLOOK_wtfact==-1)
	{	printf("\nEnter S-LOOK waisted time factor: ");
		fscanf(istream,"%d", &SLOOK_wtfact);
	}
	if (SLOOK_flag1==-2)
	{	printf("\nReset SLOOK waisted time on every min/max ? (y/n) ");
		fscanf(istream," \n\t%c", &c1); SLOOK_flag1=-(c1=='y');
	}
	if (BOPATH_wthresh==-1)
	{	printf("\nEnter BO-PATH waisted time threshold: ");
		fscanf(istream,"%d",&BOPATH_wthresh);
	}
	while (BOPATH_maxbnd<1)
	{	printf("\nEnter BO-PATH maximum A* depth (Min=1): ");
		fscanf(istream,"%d",&BOPATH_maxbnd);
		if (BOPATH_maxbnd<1) printf(">> Too low, Minimum = 1");
	}
	BOPATH_maxbnd--;
	if (BOPATH_wresetfl==-2)
	{	printf("\nReset BO-PATH waisted time on every request in path ? (y/n) ");
		fscanf(istream," \n\t%c", &c1); BOPATH_wresetfl=-(c1=='y');
	}
	if (BOPATH_swap==-2)
	{	printf("\nUse BO-PATH min/max swaping trick ? (y/n) ");
		fscanf(istream," \n\t%c", &c1); BOPATH_swap=-(c1=='y');
	}

	// Preparing queues
	q=malloc(sizeof(qreq)*(total_requests+1));			// For FCFS
	q2=malloc(sizeof(qreq)*(total_requests+1));			// For SSTF
	q2[total_requests].cyl= 5*MAXCYLINDER;
	q3=malloc(sizeof(qreq)*(total_requests+1));			// For SCAN
	q3[total_requests].cyl= 5*MAXCYLINDER;
	q4=malloc(sizeof(qreq)*(total_requests+1));			// For C-SCAN
	q4[total_requests].cyl= 5*MAXCYLINDER;
	q5=malloc(sizeof(qreq)*(total_requests+1));			// For LOOK
	q5[total_requests].cyl= 5*MAXCYLINDER;
	q6=malloc(sizeof(qreq)*(total_requests+1));			// For C-LOOK
	q6[total_requests].cyl= 5*MAXCYLINDER;

	// Disable FCFS option
	if (FCFS_nexti==-2) FCFS_nexti=total_requests+10;
	
	#ifdef RLATENCY
		BPTTIME = BPT*BLOCKTIME; cu_bltime=0;

		// Choosing Rotational latency optimization method 
		if (rlat_method==1) 
		{	get_SSTF_nexti = &get_SSTF_nexti_rlat;
			get_SCAN_nexti = &get_SCAN_nexti_rlat;
			get_cyl_nexti = &get_cyl_nexti_same;
		} else  // =2
		{	get_SSTF_nexti = &get_SSTF_nexti_nor;
			get_SCAN_nexti = &get_SCAN_nexti_nor;
			get_cyl_nexti = &get_cyl_nexti_rlat;
		} 
	#else
		get_SSTF_nexti = &get_SSTF_nexti_nor;
		get_SCAN_nexti = &get_SCAN_nexti_nor;
	#endif

	// Start Simulation
	peek_count = GetTickCount();
	poisson_e = exp(-arrival_r * time_unit); 
	printf("\nSimulation Started....\n");
	while ((rcount<total_requests) || (SSTF_cleft>0) || (FCFS_nexti<=last) || (SCAN_cleft>0) || (CSCAN_cleft>0) || (LOOK_cleft>0) || (CLOOK_cleft>0) || (SLOOK_cleft>0) || (BOPATH_cleft>0))
	{	int num_req;
		double d1;

		// Checking if caller wants to stop the simulation
		if ((GetTickCount()>peek_count+200) || GetTickCount()<peek_count) if (PeekMessage(&msg1,NULL,WM_STOP,WM_STOP,PM_REMOVE)) goto EXIT_MAIN; else peek_count=GetTickCount();
		//if (PeekMessage(&msg1,NULL,WM_STOP,WM_STOP,PM_REMOVE)) goto EXIT_MAIN;

		// Choosing number of req. to be generated
		if (!manual)
		{	
			// Possion Random Variable Generator.
			if (arrival_r!=-1)
			{	
				if (skip_idle && !SSTF_cleft && FCFS_nexti>last && !SCAN_cleft && !CSCAN_cleft && !LOOK_cleft && !CLOOK_cleft && !SLOOK_cleft && !BOPATH_cleft && CSCAN_head!=MAXCYLINDER && CSCAN_direc!=-1)
				{	num_req=0; d1 = poisson_e + ((double) genrand()/GENRAND_MAX)*(1-poisson_e);
				}
				else { num_req=-1; d1=1; }
				do 
				{	num_req++;
					d1 = d1 * ((double) genrand()/GENRAND_MAX); 
				}
				while (d1>=poisson_e);
			}
			
			// Generic Random Variable Generator.
			else if (max_req_pert)
			{	if ((genrand()<=req_prob*(GENRAND_MAX+1)-1) || (skip_idle && !SSTF_cleft && FCFS_nexti>last && !SCAN_cleft && !CSCAN_cleft && !LOOK_cleft && !CLOOK_cleft && !SLOOK_cleft && !BOPATH_cleft && CSCAN_head!=MAXCYLINDER && CSCAN_direc!=-1))
				{	num_req = (int) (((double) rand()/RAND_MAX) * (max_req_pert+1));
					if (num_req>max_req_pert) num_req=max_req_pert;
				}
				else num_req = 0;
			} else num_req=total_requests;
		
		} else  //Maual
		{	num_req=0;
			if (qmi<=qmlast) { qmi++; qmi2=qmi; }
			while (qm[qmi]>=0)
			{	num_req++; qmi++;
			}
		}
		
		rcount = rcount + num_req; 
		if (rcount > total_requests)
		{	num_req = num_req + total_requests - rcount;
			rcount = total_requests;
		}

		// Generating requests
		for (n=1; n<=num_req; n++)
		{	last++, SSTF_last++; SSTF_cleft++; SCAN_last++; SCAN_cleft++; CSCAN_last++; CSCAN_cleft++; 
			LOOK_last++; LOOK_cleft++; CLOOK_last++; CLOOK_cleft++;
			if (manual)
			{	q[last].cyl=qm[qmi2]; 
				#ifdef RLATENCY
					// Set Block number (still not maped to block time)
					q[last].bltime = qb[qmi2];
				#endif
				qmi2++;
			} else 
			{	q[last].cyl = (int) (((double) rand()/RAND_MAX) * (MAXCYLINDER+1));
				#ifdef RLATENCY
					q[last].bltime = (int) (((double) genrand()/GENRAND_MAX) * BPT);
				#endif
			}
			
			// Fixing out of bounds values. Very unlikely to happen
			if (q[last].cyl>MAXCYLINDER) q[last].cyl=MAXCYLINDER;
			#ifdef RLATENCY
				if (q[last].bltime=BPT) q[last].bltime = BPT_1;
			
				// Map the block number to it's block time.
				q[last].bltime *= BLOCKTIME;
			#endif

			// Setting the new request for all the Qs
			//if ((insort(q2,last-1,q[last].cyl)<=SSTF_nexti) && (SSTF_nexti>-1)) SSTF_nexti++;
			q2[SSTF_last].cyl=q[last].cyl; q3[SCAN_last].cyl=q[last].cyl; q4[CSCAN_last].cyl=q[last].cyl; q5[LOOK_last].cyl=q[last].cyl; q6[CLOOK_last].cyl=q[last].cyl; 
			#ifdef RLATENCY
				q2[SSTF_last].bltime=q[last].bltime; q3[SCAN_last].bltime=q[last].bltime; q4[CSCAN_last].bltime=q[last].bltime; q5[LOOK_last].bltime=q[last].bltime; q6[CLOOK_last].bltime=q[last].bltime;  
			#endif



			#ifdef RLATENCY
				if (rlat_method==1)
				{	// Fixing SCAN nexti
					if (fix_nexti_rlat(q3,SCAN_last,SCAN_head,SCAN_direc,&SCAN_nexti,SCAN_headmv)==2 && SCAN_insrv) 
						SCAN_insrv = SRVTIME(SCAN,q3);	

					// Fixing C-SCAN nexti
					if (CSCAN_direc==1)
						if (fix_nexti_rlat(q4,CSCAN_last,CSCAN_head,CSCAN_direc,&CSCAN_nexti,CSCAN_headmv)==2 && CSCAN_insrv) 
							CSCAN_insrv = SRVTIME(CSCAN,q4);

					// Fixing LOOK nexti
					if (fix_nexti_rlat(q5,LOOK_last,LOOK_head,LOOK_direc,&LOOK_nexti,LOOK_headmv)==2 && LOOK_insrv) 
						LOOK_insrv = SRVTIME(LOOK,q5);

					// Fixing C-LOOK nexti
					if (CLOOK_direc==1)
						if (fix_nexti_rlat(q6,CLOOK_last,CLOOK_head,1,&CLOOK_nexti,CLOOK_headmv)==2 && CLOOK_insrv)
							CLOOK_insrv = SRVTIME(CLOOK,q6); else ;
					else if (fix_nexti_rlat(q6,CLOOK_last,STARTCYL,1,&CLOOK_nexti,CLOOK_headmv)==2 && CLOOK_insrv)
							CLOOK_insrv = SRVTIME(CLOOK,q6);

				} else 
			#endif
				{	// Fixing SCAN nexti
					fix_nexti(q3,SCAN_last,SCAN_head,SCAN_direc,&SCAN_nexti);
				
					// Fixing C-SCAN nexti
					if (CSCAN_direc==1)
						fix_nexti(q4,CSCAN_last,CSCAN_head,CSCAN_direc,&CSCAN_nexti);

					// Fixing LOOK nexti
					fix_nexti(q5,LOOK_last,LOOK_head,LOOK_direc,&LOOK_nexti);

					// Fixing C-LOOK nexti
					if (CLOOK_direc==1)
						fix_nexti(q6,CLOOK_last,CLOOK_head,1,&CLOOK_nexti);
					else fix_nexti(q6,CLOOK_last,STARTCYL,1,&CLOOK_nexti);

				}

		}


		// FCFS
again2:	if (FCFS_headmv)
		{	FCFS_headmv--;	
			if (rcount==total_requests)
			{	FCFS_wait = FCFS_wait + (last-FCFS_nexti+1)*FCFS_headmv; 
				FCFS_headmv=0;
			}
		}
		else if (FCFS_insrv)
		{	FCFS_insrv--; if (!FCFS_insrv) FCFS_nexti++; else NOWAIT(FCFS);
		} 
		else
		{	if (FCFS_nexti<=last)
			{	n = q[FCFS_nexti].cyl-FCFS_head;
				if (n!=0)
				{	n=n/abs(n); FCFS_head = FCFS_head + n; FCFS_seek=FCFS_seek+CYLSEEK; FCFS_headmv=CYLSEEK-1;
				} else { FCFS_insrv = SRVTIME(FCFS,q); update_progress_bar(); goto again2; }
			}
		}
		FCFS_wait = FCFS_wait+last-FCFS_nexti+1; 

		
		//if ((rcount==total_requests) && SSTF_cleft==3) { printf("\n^^^%d\n",SSTF_cleft); scanf("%d",&n); }
		//if (SSTF_cleft<0) { printf("\n***%d\n",SSTF_cleft); scanf("%d",&n); }


		//SSTF
again1:	if (SSTF_headmv)
			SSTF_headmv--;
		else if (SSTF_insrv)
		{	//if ((SSTF_insrv==BLOCKTIME) && cu_bltime==q2[SSTF_nexti].bltime) printf("*");
			SSTF_insrv--; if (!SSTF_insrv) { q2[SSTF_nexti].cyl=q2[SSTF_last].cyl; SSTF_last--; SSTF_cleft--; SSTF_nexti = -1; } else NOWAIT(SSTF);
		} 
		else 
		{	if (SSTF_nexti==-1) SSTF_nexti = get_SSTF_nexti(q2,SSTF_last,SSTF_head);
			if (SSTF_nexti>-1)
			{	n = q2[SSTF_nexti].cyl-SSTF_head;
				if (n!=0)
				{	n=n/abs(n); SSTF_head = SSTF_head + n; SSTF_seek=SSTF_seek+CYLSEEK; SSTF_headmv=CYLSEEK-1;
				} else 
				{	
					#ifdef RLATENCY
						// If rlat_method=2, this will search for the nearest request in current cylinder.
						SSTF_nexti = get_cyl_nexti(q2,SSTF_last, SSTF_nexti);
					#endif
					SSTF_insrv = SRVTIME(SSTF,q2); update_progress_bar(); goto again1;
				}
			}
		}
		SSTF_wait = SSTF_wait+SSTF_cleft;


		//SCAN
again3:	if (SCAN_headmv)
			SCAN_headmv--;
		else if (SCAN_insrv)
		{	SCAN_insrv--; if (!SCAN_insrv) { q3[SCAN_nexti].cyl=q3[SCAN_last].cyl; SCAN_last--; SCAN_cleft--; SCAN_nexti = -1; } else NOWAIT(SCAN);
		} 
		else 
		{	if (SCAN_nexti==-1) SCAN_nexti = get_SCAN_nexti(q3,SCAN_last,SCAN_head,SCAN_direc);
			if (SCAN_nexti>-1)
				if (q3[SCAN_nexti].cyl==SCAN_head)
				{ 
					#ifdef RLATENCY
						// If rlat_method=2, this will search for the nearest request in current cylinder.
						SCAN_nexti = get_cyl_nexti(q3,SCAN_last, SCAN_nexti);
					#endif
					SCAN_insrv = SRVTIME(SCAN,q3); update_progress_bar(); goto again3;
				}
			if (SCAN_cleft) 
			{	if (SCAN_head==STARTCYL && SCAN_direc==-1) SCAN_direc=1;
				else if (SCAN_head==MAXCYLINDER && SCAN_direc==1) SCAN_direc=-1;
				SCAN_head = SCAN_head + SCAN_direc; SCAN_seek=SCAN_seek+CYLSEEK; SCAN_headmv=CYLSEEK-1;
			}
		}
		SCAN_wait = SCAN_wait+SCAN_cleft;


		//C-SCAN
again4:	if (CSCAN_headmv)
			CSCAN_headmv--;
		else if (CSCAN_insrv)
		{	CSCAN_insrv--; if (!CSCAN_insrv) { q4[CSCAN_nexti].cyl=q4[CSCAN_last].cyl; CSCAN_last--; CSCAN_cleft--; CSCAN_nexti = -1; } else NOWAIT(CSCAN);
		} 
		else 
		{	if (CSCAN_direc==1)
			{	if (CSCAN_nexti==-1) CSCAN_nexti = get_SCAN_nexti(q4,CSCAN_last,CSCAN_head,CSCAN_direc);
				if (CSCAN_nexti>-1)
					if (q4[CSCAN_nexti].cyl==CSCAN_head)
					{ 
						#ifdef RLATENCY
							// If rlat_method=2, this will search for the nearest request in current cylinder.
							CSCAN_nexti = get_cyl_nexti(q4,CSCAN_last, CSCAN_nexti);
						#endif
						CSCAN_insrv = SRVTIME(CSCAN,q4); update_progress_bar(); goto again4;
					}
				if (CSCAN_cleft || CSCAN_head==MAXCYLINDER) 
				{	if (CSCAN_head==MAXCYLINDER) CSCAN_direc=-1;
					CSCAN_head = CSCAN_head + CSCAN_direc; CSCAN_seek=CSCAN_seek+CYLSEEK; CSCAN_headmv=CYLSEEK-1;
				}
			}
			else 
			{	if (CSCAN_head==STARTCYL) { CSCAN_direc=1; goto again4; }
				CSCAN_head = CSCAN_head + CSCAN_direc; CSCAN_seek=CSCAN_seek+CYLSEEK; CSCAN_headmv=CYLSEEK-1;
			}
		}
		CSCAN_wait = CSCAN_wait+CSCAN_cleft;


		//LOOK
again5:	if (LOOK_headmv)
			LOOK_headmv--;
		else if (LOOK_insrv)
		{	LOOK_insrv--; if (!LOOK_insrv) { q5[LOOK_nexti].cyl=q5[LOOK_last].cyl; LOOK_last--; LOOK_cleft--; LOOK_nexti = -1; } else NOWAIT(LOOK);
		} 
		else
		{	if (LOOK_nexti==-1) LOOK_nexti = get_SCAN_nexti(q5,LOOK_last,LOOK_head,LOOK_direc);
			if (LOOK_nexti>-1)
				if (q5[LOOK_nexti].cyl==LOOK_head)
				{ 
					#ifdef RLATENCY
						// If rlat_method=2, this will search for the nearest request in current cylinder.
						LOOK_nexti = get_cyl_nexti(q5,LOOK_last, LOOK_nexti);
					#endif
					LOOK_insrv = SRVTIME(LOOK,q5); update_progress_bar(); goto again5;
				}
				else
					LOOK_direc=-LOOK_direc;
			if (LOOK_cleft) 
			{	LOOK_direc=-LOOK_direc;
				LOOK_head = LOOK_head + LOOK_direc; LOOK_seek=LOOK_seek+CYLSEEK; LOOK_headmv=CYLSEEK-1;
			}
		}
		LOOK_wait = LOOK_wait+LOOK_cleft;


		//C-LOOK
again6:	if (CLOOK_headmv)
			CLOOK_headmv--;
		else if (CLOOK_insrv)
		{	CLOOK_insrv--; if (!CLOOK_insrv) { q6[CLOOK_nexti].cyl=q6[CLOOK_last].cyl; CLOOK_last--; CLOOK_cleft--; CLOOK_nexti = -1; } else NOWAIT(CLOOK);
		} 
		else
		{	if (CLOOK_direc==1)
			{	if (CLOOK_nexti==-1) CLOOK_nexti = get_SCAN_nexti(q6,CLOOK_last,CLOOK_head,CLOOK_direc);
				if (CLOOK_nexti>-1)
					if (q6[CLOOK_nexti].cyl==CLOOK_head)
					{
						#ifdef RLATENCY
							// If rlat_method=2, this will search for the nearest request in current cylinder.
							CLOOK_nexti = get_cyl_nexti(q6,CLOOK_last, CLOOK_nexti);
						#endif
						CLOOK_insrv = SRVTIME(CLOOK,q6); update_progress_bar(); goto again6; 
					}
					else
						CLOOK_direc=-CLOOK_direc;
				if (CLOOK_cleft) 
				{	CLOOK_direc=-CLOOK_direc;
					CLOOK_head = CLOOK_head + CLOOK_direc; CLOOK_seek=CLOOK_seek+CYLSEEK; CLOOK_headmv=CYLSEEK-1;
				}
			}
			else 
			{	if (CLOOK_nexti==-1) CLOOK_nexti = get_SCAN_nexti(q6,CLOOK_last,STARTCYL,1);
				if (CLOOK_nexti==-1 || q6[CLOOK_nexti].cyl>=CLOOK_head) { CLOOK_direc=1; goto again6; }
				else { CLOOK_head--; CLOOK_seek=CLOOK_seek+CYLSEEK; CLOOK_headmv=CYLSEEK-1; }
			}
		}
		CLOOK_wait = CLOOK_wait+CLOOK_cleft;


		
		#ifdef RLATENCY
			// Rotate disk by a time unit. If block zero is reached reset time to 0
			cu_bltime++; if (cu_bltime>=BPTTIME) cu_bltime=0;
		#endif

	} // Main Simulation Loop

	if (FCFS_nexti==total_requests+10) FCFS_wait=0;

	a=(rcount-1)*BLOCKTIME;
	if (sstream!=NULL)
	{	if (scolumn!=NULL) fprintf(sstream,scolumn); fprintf(sstream,",");
		fprintf(sstream,"%d,",FCFS_seek); 
		fprintf(sstream,"%d,",SSTF_seek); 
		fprintf(sstream,"%d,",SCAN_seek); 
		fprintf(sstream,"%d,",CSCAN_seek);
		fprintf(sstream,"%d,",LOOK_seek);
		fprintf(sstream,"%d,",CLOOK_seek);
		fprintf(sstream,"%d,",SLOOK_seek);
		fprintf(sstream,"%d\n",BOPATH_seek);
		fclose(sstream);
	}
	if (wstream!=NULL)
	{	if (wcolumn!=NULL) fprintf(wstream,wcolumn); fprintf(wstream,",");
		fprintf(wstream,"%g,",(double) FCFS_wait / rcount);
		fprintf(wstream,"%g,",(double) SSTF_wait / rcount); 
		fprintf(wstream,"%g,",(double) SCAN_wait / rcount); 
		fprintf(wstream,"%g,",(double) CSCAN_wait / rcount);
		fprintf(wstream,"%g,",(double) LOOK_wait / rcount); 
		fprintf(wstream,"%g,",(double) CLOOK_wait / rcount); 
		fprintf(wstream,"%g,",(double) SLOOK_wait / rcount);
		fprintf(wstream,"%g\n",(double) BOPATH_wait / rcount);
		fclose(wstream);
	}
	
	if (sstream==NULL && wstream==NULL)
	{	printf("\nFCFS total seek = %d", FCFS_seek);
		printf("\nFCFS total seek (including block time) = %d", FCFS_seek + a);
		printf("\nFCFS average waiting time = %f\n", (double) FCFS_wait / rcount);

		printf("\nSSTF total seek = %d", SSTF_seek);
		printf("\nSSTF total seek (including block time) = %d", SSTF_seek + a);
		printf("\nSSTF average waiting time = %f\n", (double) SSTF_wait / rcount);

		printf("\nSCAN total seek = %d", SCAN_seek);
		printf("\nSCAN total seek (including block time) = %d", SCAN_seek + a);
		printf("\nSCAN average waiting time = %f\n", (double) SCAN_wait / rcount);

		printf("\nC-SCAN total seek = %d", CSCAN_seek);
		printf("\nC-SCAN total seek (including block time) = %d", CSCAN_seek + a);
		printf("\nC-SCAN average waiting time = %f\n", (double) CSCAN_wait / rcount);

		printf("\nLOOK total seek = %d", LOOK_seek);
		printf("\nLOOK total seek (including block time) = %d", LOOK_seek + a);
		printf("\nLOOK average waiting time = %f\n", (double) LOOK_wait / rcount);

		printf("\nC-LOOK total seek = %d", CLOOK_seek);
		printf("\nC-LOOK total seek (including block time) = %d", CLOOK_seek + a);
		printf("\nC-LOOK average waiting time = %f\n", (double) CLOOK_wait / rcount);
	
		printf("\nS-LOOK total seek = %d", SLOOK_seek);
		printf("\nS-LOOK total seek (including block time) = %d", SLOOK_seek + a);
		printf("\nS-LOOK average waiting time = %f\n", (double) SLOOK_wait / rcount);

		printf("\nBO-PATH total seek = %d", BOPATH_seek);
		printf("\nBO-PATH total seek (including block time) = %d", BOPATH_seek + a);
		printf("\nBO-PATH average waiting time = %f\n", (double) BOPATH_wait / rcount);
	}

	/* printf("\n");
	for (n=0; n<=last; n++)
		printf(" %d",q[n].cyl);
	printf("\n");
	free(q2); q2=malloc(sizeof(int)*(total_requests));	
	printf("Optimal average waiting time = %g\n",(double) ASTAR_opt(q,last,head,q2,&a,BLOCKTIME,CYLSEEK,1) / rcount);
	printf("Total seek = %d\n",a);
	for (n=0; n<=last; n++)
		printf(" %d",q[q2[n]].cyl); */

EXIT_MAIN:
	if (caller_hwnd != -1) PostMessage((HWND) caller_hwnd,WM_PROGRESS,chunk_count,0);

	// Cleaning
	_fcloseall();
	free(q); free(q2); free(q3); free(q4); 
	free(q5); free(q6); free(qm); free(qb);
}


/*

int insort(qreq *q, int last, int val)
{	int i=0,n;
	while (val>q[i].cyl && i<=last) i++;
	for (n=last+1; n>i; n--) q[n].cyl=q[n-1].cyl;
	q[i]=val; return i;
}
*/
